function acs_read_warm_pixels, files, $
                               split=split, $
                               nosplit=nosplit, $
                               discard=discard, $
                               progress=progress

;+
; NAME:
;      ACS_READ_TRAILS
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD detectors.
;
; PURPOSE:
;      Read in a structure containing details of all the hot pixels and surrounding pixels.
;
;      Note that it does this in a fairly clever way.
;      A simple approach can be recreated by setting SPLIT=1. This simply reads in one file
;      after another and gradually concatenates them onto a large master catalogue. However,
;      as the master catalogue gets longer, each operation takes exponentially longer.
;      
;      Setting SPLIT to higher values attempts to concatenate together as many individual 
;      catalogues as possible while they are small, then recursively build up the full
;      catalogue by sequantially piecing together these small pieces. The number of 
;      concatenation operationsis the same, but this approach optimises the number of them 
;      that involve only small arrays.
;
; INPUTS:
;      FILES    - String array containing absolute file names.
;
; OPTIONAL INPUTS:
;      SPLIT    - Subdivide the list of catalogues into this number of pieces at each step.
;                 e.g. SPLIT=2 will halve and halve and halve a catalogue until it is too small
;                 to be split. It then starts building these chunks as the smallest elements.
;      PROGRESS - This variable is used for internal reference only.
;                 DO NOT SET IT EXTERNALLY OR TRY TO RECOVER ITS CONTENTS.
;
; KEYWORD PARAMETERS:
;      DISCARD  - Discard warm pixels that are not frequently represented in the catalogue,
;                 particularly to avoid picking peaks in read noise at the faint end.
;
; OUTPUTS:
;      PIXELS   - Returns a structure containing all the necessary information.
;
; MODIFICATION HISTORY:
;      Mar 09 - RM added DISCARD option. Now even more self-recursive.
;      Jan 08 - Richard Massey felt smug about writing a self-recursive routine.
;-

; Initialise variables
if not keyword_set(files) then message,"Input file names not specified!"
n_files=n_elements(files)
if not keyword_set(split) then split=2
if not keyword_set(progress) then progress={progress:0.,increment:1./(n_files),recursion_depth:1,max_recursion_depth:1} else begin
  progress.recursion_depth=progress.recursion_depth+1
  progress.max_recursion_depth=progress.max_recursion_depth>progress.recursion_depth
endelse

if keyword_set(discard) then begin
  amplifiers=strmid(files,0,/REVERSE_OFFSET)
  reorder=sort(amplifiers)
  amp_break=uniq(amplifiers[reorder])
  amplifiers=amplifiers[reorder[amp_break]]
  for j=0,n_elements(amplifiers)-1 do begin
    amp_files=(files[reorder])[([-1,amp_break])[j]+1:([-1,amp_break])[j+1]]
    pixels=acs_read_warm_pixels(amp_files, split=split, progress=progress)
    message,/info,"Discarding pixels not found to be warm in at least 1/"+strtrim(discard,2)+" of the exposures"
    pixel_locations=bytarr(max(pixels.x)+1,max(pixels.y)+1)
    for i=0ULL,n_elements(pixels)-1 do pixel_locations[pixels[i].x,pixels[i].y]=temporary(pixel_locations[pixels[i].x,pixels[i].y])+1B
    good=where(pixel_locations[pixels.x,pixels.y] ge n_elements(amp_files)/discard)
    pixels=pixels[good]
    amp_number=(byte(strupcase(amplifiers[j]))-65)[0]
    if amp_number lt 0 or amp_number gt 6 or max(pixels.x) ge 5000 then message,/INFO,"Numerical imprecision in pixel x coordinates!"
    pixels.x=temporary(pixels.x)+5000*amp_number ; flag which amplifier they are from with a hack
    delvarx,pixel_locations
    if not keyword_set(all_pixels) then all_pixels=pixels else all_pixels=[all_pixels,temporary(pixels)]
  endfor
endif else begin


  ; Loop through files, reading them in one at a time
  if n_files ge 2*split and split gt 1 then begin
    split_points=indgen(split+1)*(n_files)/split
    all_pixels=acs_read_warm_pixels(files[split_points[0]:split_points[1]-1], split=split, progress=progress)
    for i=1,split-1 do begin
      pixels=acs_read_warm_pixels(files[split_points[i]:split_points[i+1]-1], split=split, progress=progress)
      all_pixels=[all_pixels,temporary(pixels)]
    endfor
  endif else begin
    for i=0,n_files-1 do begin
      message,/info,"("+strtrim(round(progress.progress*100),2)+"%) Reading file "+files[i]+"_pixels.sav"
      progress.progress=progress.progress+progress.increment
      restore,files[i]+"_pixels.sav"
      ;pixels=create_struct(pixels,"AMPLIFIER",replicate(strmid(files[i],0,/REVERSE_OFFSET),n_elements(pixels.x)))
      if not keyword_set(all_pixels) then all_pixels=pixels else all_pixels=[all_pixels,temporary(pixels)]
    endfor
    ;print,round(progress.progress*100),"%...",format="(I0,A0,$)"
  endelse

endelse

; Chart progress and report back
progress.recursion_depth=progress.recursion_depth-1
if progress.recursion_depth eq 0 then begin
  message,/INFO,"(100%) Maximum recursion depth: "+strtrim(progress.max_recursion_depth,2)
  delvarx,progress
endif
return,all_pixels

end
